@@ -29,6 +29,7 @@ module System.IO.Streams.ByteString
  , giveBytes
  , giveExactly
  , takeBytes
+ , takeExactly
  , throwIfConsumesMoreThan
  , throwIfProducesMoreThan
 
@@ -57,26 +58,18 @@ import qualified Data.ByteString.Lazy.Char8        as L
 import qualified Data.ByteString.Unsafe            as S
 import           Data.Char                         (isSpace)
 import           Data.Int                          (Int64)
-import           Data.IORef                        (IORef, newIORef,
-                                                    readIORef, writeIORef)
+import           Data.IORef                        (IORef, newIORef, readIORef, writeIORef)
 import           Data.Time.Clock.POSIX             (getPOSIXTime)
 import           Data.Typeable                     (Typeable)
-
-import           Prelude                           hiding (lines, read,
-                                                    takeWhile, unlines,
-                                                    unwords, words)
+import           Prelude                           hiding (read, lines, unlines, words, unwords)
 ------------------------------------------------------------------------------
-import           System.IO.Streams.Combinators     (filterM, intersperse,
-                                                    outputFoldM)
-import           System.IO.Streams.Internal        (InputStream (..),
-                                                    OutputStream,
-                                                    makeInputStream,
-                                                    makeOutputStream, read,
-                                                    unRead, write)
+import           System.IO.Streams.Combinators     (filterM, intersperse, outputFoldM)
+import           System.IO.Streams.Internal        (InputStream (..), OutputStream, makeInputStream, makeOutputStream, read, unRead, write)
 import           System.IO.Streams.Internal.Search (MatchInfo (..), search)
 import           System.IO.Streams.List            (fromList, writeList)
-------------------------------------------------------------------------------
 
+
+------------------------------------------------------------------------------
 {-# INLINE modifyRef #-}
 modifyRef :: IORef a -> (a -> a) -> IO ()
 modifyRef ref f = do
@@ -215,7 +208,29 @@ countOutput = outputFoldM f 0
 takeBytes :: Int64                        -- ^ maximum number of bytes to read
           -> InputStream ByteString       -- ^ input stream to wrap
           -> IO (InputStream ByteString)
-takeBytes k0 src = do
+takeBytes k0 = takeBytes' k0 (return Nothing)
+{-# INLINE takeBytes #-}
+
+
+------------------------------------------------------------------------------
+-- | Like @Streams.'takeBytes'@, but throws 'ReadTooShortException' when
+-- there aren't enough bytes present on the source.
+takeExactly :: Int64                        -- ^ number of bytes to read
+            -> InputStream ByteString       -- ^ input stream to wrap
+            -> IO (InputStream ByteString)
+takeExactly k0 = takeBytes' k0 (throwIO $ ReadTooShortException k0)
+{-# INLINE takeExactly #-}
+
+
+------------------------------------------------------------------------------
+-- Helper for the two above.
+takeBytes' :: Int64
+           -> IO (Maybe ByteString)
+           -- ^ What to do if the input ends before having consumed the
+           -- right amount of bytes.
+           -> InputStream ByteString
+           -> IO (InputStream ByteString)
+takeBytes' k0 h src = do
     kref <- newIORef k0
     return $! InputStream (prod kref) (pb kref)
   where
@@ -223,7 +238,7 @@ takeBytes k0 src = do
         k <- readIORef kref
         if k <= 0
            then return Nothing
-           else read src >>= maybe (return Nothing) (chunk k)
+           else read src >>= maybe h (chunk k)
       where
         chunk k s = do
             let l  = fromIntegral $ S.length s
@@ -239,6 +254,7 @@ takeBytes k0 src = do
     pb kref s = do
         modifyRef kref (+ (fromIntegral $ S.length s))
         unRead s src
+{-# INLINE takeBytes' #-}
 
 
 ------------------------------------------------------------------------------
@@ -404,8 +420,9 @@ instance Exception TooManyBytesWrittenException
 
 
 ------------------------------------------------------------------------------
--- | Thrown by 'readExactly' when not enough bytes were available on the input.
-data ReadTooShortException = ReadTooShortException Int deriving (Typeable)
+-- | Thrown by 'readExactly' and 'takeExactly' when not enough bytes were
+-- available on the input.
+data ReadTooShortException = ReadTooShortException Int64 deriving (Typeable)
 
 instance Show ReadTooShortException where
     show (ReadTooShortException x) = "Short read, expected " ++ show x
@@ -501,7 +518,7 @@ readExactly n input = go id n
     go !dl 0  = return $! S.concat $! dl []
     go !dl k  =
         read input >>=
-        maybe (throwIO $ ReadTooShortException n)
+        maybe (throwIO $ ReadTooShortException (fromIntegral n))
               (\s -> do
                  let l = S.length s
                  if l >= k
