@@ -2,6 +2,7 @@
 
 {-# LANGUAGE BangPatterns       #-}
 {-# LANGUAGE DeriveDataTypeable #-}
+{-# LANGUAGE RankNTypes         #-}
 
 module System.IO.Streams.Combinators
  ( -- * Folds
@@ -21,9 +22,11 @@ module System.IO.Streams.Combinators
  , map
  , mapM
  , mapM_
+ , mapMaybe
  , contramap
  , contramapM
  , contramapM_
+ , contramapMaybe
 
    -- * Filter
  , filter
@@ -56,17 +59,17 @@ import           Control.Concurrent.MVar    (newMVar, withMVar)
 import           Control.Monad              (liftM, void, when)
 import           Control.Monad.IO.Class     (liftIO)
 import           Data.Int                   (Int64)
-import           Data.IORef                 (atomicModifyIORef, modifyIORef, newIORef, readIORef, writeIORef)
+import           Data.IORef                 (IORef, atomicModifyIORef, modifyIORef, newIORef, readIORef, writeIORef)
 import           Data.Maybe                 (isJust)
 import           Prelude                    hiding (all, any, drop, filter, map, mapM, mapM_, maximum, minimum, read, take, unzip, zip, zipWith)
 ------------------------------------------------------------------------------
-import           System.IO.Streams.Internal (InputStream (..), OutputStream, fromGenerator, makeInputStream, makeOutputStream, read, unRead, write, yield)
+import           System.IO.Streams.Internal (InputStream (..), OutputStream (..), fromGenerator, makeInputStream, makeOutputStream, read, unRead, write, yield)
 
 
 ------------------------------------------------------------------------------
 -- | A side-effecting fold over an 'OutputStream', as a stream transformer.
 --
--- The IO action returned by 'outputFoldM' can be used to fetch the updated
+-- The IO action returned by 'outputFoldM' can be used to fetch and reset the updated
 -- seed value. Example:
 --
 -- @
@@ -83,8 +86,8 @@ outputFoldM :: (a -> b -> IO a)           -- ^ fold function
             -> a                          -- ^ initial seed
             -> OutputStream b             -- ^ output stream
             -> IO (OutputStream b, IO a)  -- ^ returns a new stream as well as
-                                          -- an IO action to fetch the updated
-                                          -- seed value.
+                                          -- an IO action to fetch and reset the
+                                          --  updated seed value.
 outputFoldM f initial stream = do
     ref <- newIORef initial
     os  <- makeOutputStream (wr ref)
@@ -104,7 +107,7 @@ outputFoldM f initial stream = do
 ------------------------------------------------------------------------------
 -- | A side-effecting fold over an 'InputStream', as a stream transformer.
 --
--- The IO action returned by 'inputFoldM' can be used to fetch the updated seed
+-- The IO action returned by 'inputFoldM' can be used to fetch and reset the updated seed
 -- value. Example:
 --
 -- @
@@ -119,8 +122,8 @@ inputFoldM :: (a -> b -> IO a)          -- ^ fold function
            -> a                         -- ^ initial seed
            -> InputStream b             -- ^ input stream
            -> IO (InputStream b, IO a)  -- ^ returns a new stream as well as an
-                                        -- IO action to fetch the updated seed
-                                        -- value.
+                                        -- IO action to fetch and reset the
+                                        -- updated seed value.
 inputFoldM f initial stream = do
     ref <- newIORef initial
     is  <- makeInputStream (rd ref)
@@ -368,6 +371,33 @@ mapM_ f s = makeInputStream $ do
     return mb
 
 
+------------------------------------------------------------------------------
+-- | A version of map that discards elements
+--
+-- @mapMaybe f s@ passes all output from @s@ through the function @f@ and
+-- discards elements for which @f s@ evaluates to 'Nothing'.
+--
+-- Example:
+--
+-- @
+-- ghci> Streams.'System.IO.Streams.fromList' [Just 1, None, Just 3] >>=
+--       Streams.'mapMaybe' 'id' >>=
+--       Streams.'System.IO.Streams.toList'
+-- [1,3]
+-- @
+--
+-- /Since: 1.2.1.0/
+mapMaybe :: (a -> Maybe b) -> InputStream a -> IO (InputStream b)
+mapMaybe f src = makeInputStream g
+  where
+    g = do
+      s <- read src
+      case s of
+        Nothing -> return Nothing
+        Just x ->
+          case f x of
+            Nothing -> g
+            y -> return y
 ------------------------------------------------------------------------------
 -- | Contravariant counterpart to 'map'.
 --
@@ -416,6 +446,24 @@ contramapM_ f s = makeOutputStream $ \mb -> do
     write mb s
 
 
+------------------------------------------------------------------------------
+-- | Contravariant counterpart to 'contramapMaybe'.
+--
+-- @contramap f s@ passes all input to @s@ through the function @f@.
+-- Discards all the elements for which @f@ returns 'Nothing'.
+--
+-- /Since: 1.2.1.0/
+--
+contramapMaybe :: (a -> Maybe b) -> OutputStream b -> IO (OutputStream a)
+contramapMaybe f s = makeOutputStream $ g
+    where
+      g Nothing = write Nothing s
+      g (Just a) =
+        case f a of
+          Nothing -> return ()
+          x -> write x s
+
+
 ------------------------------------------------------------------------------
 -- | Drives an 'InputStream' to end-of-stream, discarding all of the yielded
 -- values.
@@ -606,33 +654,43 @@ filterOutputM p output = makeOutputStream chunk
 -- stream.
 --
 -- Access to the original stream is thread safe, i.e. guarded by a lock.
-unzip :: InputStream (a, b) -> IO (InputStream a, InputStream b)
+unzip :: forall a b . InputStream (a, b) -> IO (InputStream a, InputStream b)
 unzip os = do
     lock <- newMVar $! ()
     buf1 <- newIORef id
     buf2 <- newIORef id
 
-    is1  <- makeInputStream $ src lock id buf1 buf2
-    is2  <- makeInputStream $ src lock twist buf2 buf1
+    is1  <- makeInputStream $ src1 lock buf1 buf2
+    is2  <- makeInputStream $ src2 lock buf1 buf2
 
     return (is1, is2)
 
   where
-    twist (a, b) = (b, a)
-
-    src lock proj myBuf theirBuf = withMVar lock $ const $ do
-        dl <- readIORef myBuf
+    twist (a,b) = (b,a)
 
+    src1 lock aBuf bBuf = withMVar lock $ const $ do
+        dl <- readIORef aBuf
         case dl [] of
-          []     -> more
-          (x:xs) -> writeIORef myBuf (xs++) >> (return $! Just x)
-      where
-        more = read os >>=
-               maybe (return Nothing)
-                     (\x -> do
-                          let (a, b) = proj x
-                          modifyIORef theirBuf (. (b:))
-                          return $! Just a)
+          []     -> more os id bBuf
+          (x:xs) -> writeIORef aBuf (xs++) >> (return $! Just x)
+
+    src2 lock aBuf bBuf = withMVar lock $ const $ do
+        dl <- readIORef bBuf
+        case dl [] of
+          []     -> more os twist aBuf
+          (y:ys) -> writeIORef bBuf (ys++) >> (return $! Just y)
+
+    more :: forall a b x y .
+            InputStream (a,b)
+         -> ((a,b) -> (x,y))
+         -> IORef ([y] -> [y])
+         -> IO (Maybe x)
+    more origs proj buf = read origs >>=
+                          maybe (return Nothing)
+                                (\x -> do
+                                    let (a, b) = proj x
+                                    modifyIORef buf (. (b:))
+                                    return $! Just a)
 
 
 ------------------------------------------------------------------------------
@@ -755,7 +813,7 @@ ignore k output = newIORef k >>= makeOutputStream . chunk
 -- /Since: 1.0.1.0/
 --
 ignoreEof :: OutputStream a -> IO (OutputStream a)
-ignoreEof s = makeOutputStream f
+ignoreEof s = return $ OutputStream f
   where
     f Nothing  = return $! ()
     f x        = write x s
