@@ -4,6 +4,7 @@
 -- Library users should use the interface provided by "System.IO.Streams"
 
 {-# LANGUAGE BangPatterns               #-}
+{-# LANGUAGE CPP                        #-}
 {-# LANGUAGE DeriveDataTypeable         #-}
 {-# LANGUAGE FlexibleInstances          #-}
 {-# LANGUAGE GeneralizedNewtypeDeriving #-}
@@ -27,6 +28,7 @@ module System.IO.Streams.Internal
   , unRead
   , peek
   , write
+  , writeTo
   , atEOF
 
     -- * Building streams
@@ -61,7 +63,9 @@ module System.IO.Streams.Internal
   ) where
 
 ------------------------------------------------------------------------------
-import           Control.Applicative      (Applicative (..))
+#if !MIN_VERSION_base(4,8,0)
+import           Control.Applicative      (Applicative (..), (<$>))
+#endif
 import           Control.Concurrent       (newMVar, withMVar)
 import           Control.Exception        (throwIO)
 import           Control.Monad            (when)
@@ -148,6 +152,15 @@ write = flip _write
 {-# INLINE write #-}
 
 
+------------------------------------------------------------------------------
+-- | Flipped version of 'write'.
+--
+-- /Since: 1.3.0.0./
+writeTo :: OutputStream a -> Maybe a -> IO ()
+writeTo = _write
+{-# INLINE writeTo #-}
+
+
 ------------------------------------------------------------------------------
 -- | Observes the first value from an 'InputStream' without consuming it.
 --
@@ -269,8 +282,21 @@ makeInputStream m = do
 -- | Creates an 'OutputStream' from a value-consuming action.
 --
 -- (@makeOutputStream f@) runs the computation @f@ on each value fed to it.
+--
+-- Since version 1.2.0.0, 'makeOutputStream' also ensures that output streams
+-- no longer receive data once EOF is received (i.e. you can now assume that
+-- makeOutputStream will feed your function @Nothing@ at most once.)
 makeOutputStream :: (Maybe a -> IO ()) -> IO (OutputStream a)
-makeOutputStream = return . OutputStream
+makeOutputStream func = (OutputStream . go) <$> newIORef False
+  where
+    go closedRef !m = do
+        closed <- readIORef closedRef
+        if closed
+          then return $! ()
+          else do
+            when (isNothing m) $ writeIORef closedRef True
+            func m
+{-# INLINE makeOutputStream #-}
 
 
 ------------------------------------------------------------------------------
