@@ -3,21 +3,27 @@
 module System.IO.Streams.Tests.Network (tests) where
 
 ------------------------------------------------------------------------------
-import           Control.Concurrent             (forkIO, newEmptyMVar,
-                                                 putMVar, takeMVar)
-import qualified Network.Socket                 as N
-import           System.Timeout                 (timeout)
+import           Control.Concurrent                 (forkIO, newEmptyMVar, putMVar, takeMVar)
+import           Control.Monad                      (join)
+import qualified Data.ByteString.Char8              as S
+import           Data.IORef                         (atomicModifyIORef, newIORef)
+import qualified Network.Socket                     as N
+import           System.IO.Error                    (eofErrorType, mkIOError)
+import           System.Timeout                     (timeout)
 import           Test.Framework
 import           Test.Framework.Providers.HUnit
-import           Test.HUnit                     hiding (Test)
+import           Test.HUnit                         hiding (Test)
 ------------------------------------------------------------------------------
-import qualified System.IO.Streams.Internal     as Streams
-import qualified System.IO.Streams.List         as Streams
-import qualified System.IO.Streams.Network      as Streams
+import qualified System.IO.Streams.Internal         as Streams
+import qualified System.IO.Streams.Internal.Network as Streams
+import qualified System.IO.Streams.List             as Streams
 ------------------------------------------------------------------------------
+import           System.IO.Streams.Tests.Common     (expectExceptionH)
 
 tests :: [Test]
-tests = [ testSocket ]
+tests = [ testSocket
+        , testSocketWithError
+        ]
 
 testSocket :: Test
 testSocket = testCase "network/socket" $
@@ -59,3 +65,24 @@ testSocket = testCase "network/socket" $
         Streams.toList is >>= flip Streams.writeList os
         N.sClose csock
         N.sClose sock
+
+testSocketWithError :: Test
+testSocketWithError = testCase "network/socket-error" $ N.withSocketsDo $ do
+    codes1 <- newIORef [ return 1
+                       , ioError $ mkIOError eofErrorType "eof" Nothing Nothing ]
+    codes2 <- newIORef [ return 1
+                       , ioError $ userError "foo" ]
+
+    (is1, _)  <- Streams.socketToStreamsWithBufferSizeImpl (rbuf codes1) 64 (error "z")
+    (Just s1) <- Streams.read is1
+    assertEqual "one byte" 1 $ S.length s1
+    Nothing   <- Streams.read is1
+
+    (is2, _)  <- Streams.socketToStreamsWithBufferSizeImpl (rbuf codes2) 64 undefined
+    (Just s2) <- Streams.read is2
+    assertEqual "one byte" 1 $ S.length s2
+    expectExceptionH $ Streams.read is2
+
+  where
+    rbuf rcodes _ _ _ = join $ atomicModifyIORef rcodes $ \codes ->
+                        (tail codes, head codes)
